﻿
Imports System
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Linq
Imports System.Linq.Expressions
Imports System.Reflection
Imports System.Dynamic
Imports System.Windows
'Imports Expression = System.Linq.Expressions.Expression


Public Class ViewModelBase
    Inherits DynamicObject
    Implements INotifyPropertyChanged
    Private ReadOnly _values As New Dictionary(Of String, Object)()
    Private ReadOnly _propertyMap As IDictionary(Of String, List(Of String))
    Private ReadOnly _methodMap As IDictionary(Of String, List(Of String))
    Private ReadOnly _commandMap As IDictionary(Of String, List(Of String))

    <AttributeUsage(AttributeTargets.Property Or AttributeTargets.Method, AllowMultiple:=True)>
    Protected Class DependsUponAttribute
        Inherits Attribute
        Private _DependencyName As String
        Public Property DependencyName() As String
            Get
                Return _DependencyName
            End Get
            Private Set(ByVal value As String)
                _DependencyName = value
            End Set
        End Property

        Public Property VerifyStaticExistence() As Boolean

        Public Sub New(ByVal propertyName As String)
            DependencyName = propertyName
        End Sub
    End Class

    Private Shared ReadOnly _isInDesignMode As Boolean

    Shared Sub New()
        ' Thanks to Laurent Bugnion for this detection code:
        ' http://geekswithblogs.net/lbugnion/archive/2009/09/05/detecting-design-time-mode-in-wpf-and-silverlight.aspx
#If SILVERLIGHT Then
			_isInDesignMode = DesignerProperties.IsInDesignTool
#Else
        Dim prop = DesignerProperties.IsInDesignModeProperty
        _isInDesignMode = CBool(DependencyPropertyDescriptor.FromProperty(prop, GetType(FrameworkElement)).Metadata.DefaultValue)
#End If
    End Sub

    Public ReadOnly Property IsInDesignMode() As Boolean
        Get
            Return _isInDesignMode
        End Get
    End Property

    Public Sub New()
        _propertyMap = MapDependencies(Of DependsUponAttribute)(Function() Me.GetType().GetProperties())
        _methodMap = MapDependencies(Of DependsUponAttribute)(Function() Me.GetType().GetMethods().Cast(Of MemberInfo)().Where(Function(method) (Not method.Name.StartsWith(CAN_EXECUTE_PREFIX))))
        _commandMap = MapDependencies(Of DependsUponAttribute)(Function() Me.GetType().GetMethods().Cast(Of MemberInfo)().Where(Function(method) method.Name.StartsWith(CAN_EXECUTE_PREFIX)))
        CreateCommands()
        VerifyDependancies()
    End Sub

    Protected Function [Get](Of T)(ByVal name As String) As T
        Return [Get](name, CType(Nothing, Object))
    End Function

    Protected Function [Get](Of T)(ByVal name As String, ByVal defaultValue As T) As T
        If _values.ContainsKey(name) Then
            Return CType(_values(name), T)
        End If
        Return defaultValue
    End Function

    Protected Function [Get](Of T)(ByVal name As String, ByVal initialValue As Func(Of T)) As T
        If _values.ContainsKey(name) Then
            Return CType(_values(name), T)
        End If

        [Set](name, initialValue())
        Return [Get](Of T)(name)
    End Function

    Function [Get](Of T)(ByVal expression As Expression(Of Func(Of T))) As T
        Return [Get](Of T)(PropertyName(expression))
    End Function

    Protected Function [Get](Of T)(ByVal expression As Expression(Of Func(Of T)), ByVal defaultValue As T) As T
        Return [Get](PropertyName(expression), defaultValue)
    End Function

    Protected Function [Get](Of T)(ByVal expression As Expression(Of Func(Of T)), ByVal initialValue As Func(Of T)) As T
        Return [Get](PropertyName(expression), initialValue:=initialValue)
    End Function

    Public Sub [Set](Of T)(ByVal name As String, ByVal value As T)
        If _values.ContainsKey(name) Then
            If _values(name) Is Nothing AndAlso value Is Nothing Then
                Return
            End If

            If _values(name) IsNot Nothing AndAlso _values(name).Equals(value) Then
                Return
            End If

            _values(name) = value
        Else
            _values.Add(name, value)
        End If

        RaisePropertyChanged(name)
    End Sub

    Protected Sub RaisePropertyChanged(ByVal name As String)
        PropertyChangedEvent.Raise(Me, name)
#If SILVERLIGHT Then
			PropertyChanged.Raise(Me, "")
#End If

        If _propertyMap.ContainsKey(name) Then
            _propertyMap(name).Each(AddressOf RaisePropertyChanged)
        End If

        ExecuteDependentMethods(name)
        FireChangesOnDependentCommands(name)
    End Sub

    Private Sub ExecuteDependentMethods(ByVal name As String)
        If _methodMap.ContainsKey(name) Then
            _methodMap(name).Each(AddressOf ExecuteMethod)
        End If
    End Sub

    Private Sub FireChangesOnDependentCommands(ByVal name As String)
        If _commandMap.ContainsKey(name) Then
            _commandMap(name).Each(AddressOf RaiseCanExecuteChangedEvent)
        End If
    End Sub

    Protected Sub [Set](Of T)(ByVal expression As Expression(Of Func(Of T)), ByVal value As T)
        [Set](PropertyName(expression), value)
    End Sub

    Private Shared Function PropertyName(Of T)(ByVal expression As Expression(Of Func(Of T))) As String
        Dim memberExpression = TryCast(expression.Body, MemberExpression)

        If memberExpression Is Nothing Then
            Throw New ArgumentException("expression must be a property expression")
        End If

        Return memberExpression.Member.Name
    End Function

    Public Overrides Function TryGetMember(ByVal binder As GetMemberBinder, <System.Runtime.InteropServices.Out()> ByRef result As Object) As Boolean
        result = [Get](Of Object)(binder.Name)

        If result IsNot Nothing Then
            Return True
        End If

        Return MyBase.TryGetMember(binder, result)
    End Function

    Public Overrides Function TrySetMember(ByVal binder As SetMemberBinder, ByVal value As Object) As Boolean
        Dim result = MyBase.TrySetMember(binder, value)
        If result Then
            Return True
        End If

        [Set](binder.Name, value)
        Return True
    End Function

    Public Event PropertyChanged As PropertyChangedEventHandler Implements INotifyPropertyChanged.PropertyChanged

    Private Sub CreateCommands()
        CommandNames.Each(Sub(name) [Set](name, New DelegateCommand(Of Object)(Sub(x) ExecuteCommand(name, x), Function(x) CanExecuteCommand(name, x))))
    End Sub

    Private Const EXECUTE_PREFIX As String = "Execute_"
    Private Const CAN_EXECUTE_PREFIX As String = "CanExecute_"

    Private ReadOnly Property CommandNames() As IEnumerable(Of String)
        Get
            Return From method In Me.GetType().GetMethods()
                   Where method.Name.StartsWith(EXECUTE_PREFIX)
                   Select method.Name.StripLeft(EXECUTE_PREFIX.Length)
        End Get
    End Property

    Private Sub ExecuteCommand(ByVal name As String, ByVal parameter As Object)
        Dim methodInfo = Me.GetType().GetMethod(EXECUTE_PREFIX & name)
        If methodInfo Is Nothing Then
            Return
        End If

        methodInfo.Invoke(Me, If(methodInfo.GetParameters().Length = 1, {parameter}, Nothing))
    End Sub

    Private Function CanExecuteCommand(ByVal name As String, ByVal parameter As Object) As Boolean
        Dim methodInfo = Me.GetType().GetMethod(CAN_EXECUTE_PREFIX & name)
        If methodInfo Is Nothing Then
            Return True
        End If

        Return CBool(methodInfo.Invoke(Me, If(methodInfo.GetParameters().Length = 1, {parameter}, Nothing)))
    End Function

    Protected Sub RaiseCanExecuteChangedEvent(ByVal canExecute_name As String)
        Dim commandName = canExecute_name.StripLeft(CAN_EXECUTE_PREFIX.Length)
        Dim command = [Get](Of DelegateCommand(Of Object))(commandName)
        If command Is Nothing Then
            Return
        End If

        command.RaiseCanExecuteChanged()
    End Sub

#If SILVERLIGHT Then
		Default Public Property Item(ByVal key As String) As Object
			Get
				Return [Get](Of Object)(key)
			End Get
			Set(ByVal value As Object)
				[Set](key, value)
			End Set
		End Property
#End If

    Private Shared Function MapDependencies(Of T As DependsUponAttribute)(ByVal getInfo As Func(Of IEnumerable(Of MemberInfo))) As IDictionary(Of String, List(Of String))
        Dim dependencyMap = getInfo().ToDictionary(Function(p) p.Name, Function(p) p.GetCustomAttributes(GetType(T), True).Cast(Of T)().Select(Function(a) a.DependencyName).ToList())

        Return Invert(dependencyMap)
    End Function

    Private Shared Function Invert(ByVal map As IDictionary(Of String, List(Of String))) As IDictionary(Of String, List(Of String))
        Dim flattened = From key In map.Keys, value In map(key)
                        Select New With {Key .Key = key, Key .Value = value}

        Dim uniqueValues = flattened.Select(Function(x) x.Value).Distinct()

        Return uniqueValues.ToDictionary(Function(x) x, Function(x) (
            From item In flattened
            Where item.Value = x
            Select item.Key).ToList())
    End Function

    Private Sub ExecuteMethod(ByVal name As String)
        Dim memberInfo = Me.GetType().GetMethod(name)
        If memberInfo Is Nothing Then
            Return
        End If

        memberInfo.Invoke(Me, Nothing)
    End Sub

    Private Sub VerifyDependancies()
        Dim methods = Me.GetType().GetMethods().Cast(Of MemberInfo)()
        Dim properties = Me.GetType().GetProperties()

        Dim propertyNames = methods.Union(properties).SelectMany(Function(method) method.GetCustomAttributes(GetType(DependsUponAttribute), True).Cast(Of DependsUponAttribute)()).Where(Function(attribute) attribute.VerifyStaticExistence).Select(Function(attribute) attribute.DependencyName)

        propertyNames.Each(AddressOf VerifyDependancy)
    End Sub

    Private Sub VerifyDependancy(ByVal propertyName As String)
        Dim [property] = Me.GetType().GetProperty(propertyName)
        If [property] Is Nothing Then
            Throw New ArgumentException("DependsUpon Property Does Not Exist: " & propertyName)
        End If
    End Sub
End Class
